import pandas as pd

from ..provider_base import ProviderBase


class Provider(ProviderBase):

    def header(self):
        """Header of data stored

        Returns:
            (dict): [unit] description per columns
        """
        return {"dmy": "[smth] description"}

    def get(self, mandatory, optional=True, missing="raise"):
        """Get data from local storage.

        Raises: if missing="raise", KeyError if either pos or some dates have not been fetch
                previously on server

        Args:
            mandatory (float): mandatory argument
            optional (bool): optional argument
            missing (str): whether to fetch missing entry from remote server or raise error

        Returns:
            (pd.DataFrame): name of columns varies
        """
        assert optional

        df = pd.DataFrame([{"index": i, "dmy": mandatory} for i in range(10)]).set_index("index")

        return df, {"index": "something", "dmy": "[-] unknown"}



import logging.handlers
import os
import time
from pathlib import Path

from flask import Flask


def init_logger():
    fmt = logging.Formatter("%(asctime)s %(levelname)s (%(name)s): %(message)s")
    fmt.converter = time.gmtime
    wng_ch = logging.StreamHandler()
    wng_ch.setLevel(logging.DEBUG)
    wng_ch.setFormatter(fmt)

    pth_log = Path("cachub_logging/journal.log")
    pth_log.parent.mkdir(parents=True, exist_ok=True)
    frot = logging.handlers.TimedRotatingFileHandler(pth_log, when="D", interval=1)
    frot.setLevel(logging.DEBUG)
    frot.setFormatter(fmt)

    logger = logging.getLogger("cachub")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(wng_ch)
    logger.addHandler(frot)


def create_app(test_config=None):
    app = Flask(__name__, instance_relative_config=True)
    app.config.from_mapping(
        SECRET_KEY='dev',
        DATABASE=os.path.join(app.instance_path, 'flaskr.sqlite'),
    )

    if test_config is None:
        # load the instance config, if it exists, when not testing
        app.config.from_pyfile('config.py', silent=True)
    else:
        # load the test config if passed in
        app.config.from_mapping(test_config)

    # ensure the instance folder exists
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    init_logger()

    @app.route('/')
    def page_home():
        logger = logging.getLogger("cachub")
        logger.info("Home")
        return 'Hello, World!'

    from . import wsgi
    app.register_blueprint(wsgi.bp)

    return app



from pathlib import Path

import pandas as pd

from ..provider_base import ProviderBase


class Provider(ProviderBase):

    def __init__(self):
        super().__init__()
        self.storage = self.storage_root() / "arbo"

    def header(self):
        """Header of data stored

        Returns:
            (dict): [unit] description per columns
        """
        return {}

    def pth_storage(self, pth, anchor):
        """Normalized name associated to given coordinates

        Args:
            pth (Path): path to local file
            anchor (str): name of directory to use as anchor to make pth relative

        Returns:
            (Pth)
        """
        return self.storage / pth.name


    def get(self, pth, anchor="work"):
        """Get local pth from previously pushed file in storage.

        Args:
            pth (Path): path to local file
            anchor (str): name of directory to use as anchor to make pth relative

        Returns:
            (pd.DataFrame): name of columns varies
        """
        if isinstance(pth, str):
            pth = Path(pth)

        df = pd.read_csv(self.pth_storage(pth), sep=";", comment="#")

        return df



import json
from datetime import datetime

import pandas as pd
import requests

# api
base_url = "https://archive-api.open-meteo.com/v1/archive"

cmap = {
    "temperature_2m": "t_air",
    "relative_humidity_2m": "rh",
    "precipitation": "rain",
    "shortwave_radiation": "ghi",
    "wind_speed_10m": "ws",
    "et0_fao_evapotranspiration": "eto",
}


def fetch(latitude, longitude, date_beg, date_end):
    """Fetch time series for given location on remote server

    Args:
        latitude (float): [deg] latitude of place
        longitude (float): [deg] longitude of place
        date_beg (datetime): First date, included
        date_end (datetime): End date, included

    Returns:
        (pd.DataFrame, dict): date, header
    """
    params = {
        "latitude": latitude,
        "longitude": longitude,
        "start_date": date_beg.date().isoformat(),
        "end_date": date_end.date().isoformat(),
        "wind_speed_unit": "ms",
        "hourly": list(cmap.keys())
    }

    # fetch
    response = requests.get(url=base_url, params=params, verify=True, timeout=30.00)

    cnt = response.content
    data = json.loads(cnt)
    if "hourly" not in data:
        raise UserWarning(data)

    df = pd.DataFrame(data["hourly"])
    df = df.rename(columns=cmap | {'time': 'date'})
    df["date"] = pd.to_datetime(df["date"])

    # format header
    units = {name_norm: data["hourly_units"][name] for name, name_norm in cmap.items()}
    assert units["ghi"] == "W/m²"
    units["ghi"] = "W.m-2"
    assert units["ws"] == "m/s"
    units["ws"] = "m.s-1"
    assert units["rain"] == "mm"
    units["rain"] = "mm.h-1"
    assert units["eto"] == "mm"
    units["eto"] = "mm.h-1"

    header = {name: f"[{unit}]" for name, unit in units.items()}
    header["date"] = "[utc] see https://open-meteo.com/en/docs/historical-weather-api"
    header["latitude"] = f"[deg] {latitude:.6f}"
    header["longitude"] = f"[deg] {longitude:.6f}"

    return df.set_index("date"), header



from pathlib import Path

from .config_management import get_config


class ProviderBase:
    """Base class for all data providers

    Each model must implement a derived class from this one

    The main objective is to encapsulate the logic of calling distant server
    to fetch data and store them sparsely on disk.
    """

    def __init__(self):
        cfg = get_config()
        self._storage_root = Path.home() / cfg["storage"]["pth"]

    def storage_root(self):
        return self._storage_root

    def get(self, *args, **kwds):
        """Data stored and associated header

        Returns:
            (Any, dict): data, [unit] description per columns
        """
        pass



"""
Handle configuration for this package install
"""
import configparser
from pathlib import Path

pth_cfg = Path.home() / ".cachub"


def get_config():
    if not pth_cfg.exists():
        config = configparser.ConfigParser()
        config['storage'] = {'pth': 'cachub/storage/'}
        with open(pth_cfg, 'w') as fhw:
            config.write(fhw)

    config = configparser.ConfigParser()
    config.read(pth_cfg)

    return config



import datetime
import functools
import logging
import os

import bcrypt
import jwt
from flask import jsonify, request

SECRET_KEY = os.environ["CACHUB_DEV"]
logger = logging.getLogger(__name__)


def hash_password(password):
    """Hash a password using bcrypt"""
    ## Convert password to bytes if it's a string
    if isinstance(password, str):
        password = password.encode('utf-8')

    ## Generate a salt and hash the password
    salt = bcrypt.gensalt()
    hashed = bcrypt.hashpw(password, salt)

    return hashed


def check_password(password, hashed):
    """Verify a password against its hash"""
    ## Convert password to bytes if it's a string
    if isinstance(password, str):
        password = password.encode('utf-8')

    ## Check if the password matches the hash
    return bcrypt.checkpw(password, hashed)


def generate_token(username, role, duration=None):
    """Generate a JWT token for the authenticated user

    Args:
        username (str): name of user
        role (str): typez of access granted to user
        duration (int|None): expiration date if not None

    Returns:

    """
    payload = {
        'username': username,
        'role': role,
    }
    if duration is not None:
        expiration = datetime.datetime.utcnow() + datetime.timedelta(minutes=duration)
        payload["exp"] = expiration

    ## Create the JWT token
    token = jwt.encode(payload, SECRET_KEY, algorithm='HS256')
    return token


def verify_token(token):
    """Verify the JWT token"""
    try:
        ## Decode and verify the token
        payload = jwt.decode(token, SECRET_KEY, algorithms=['HS256'])
        return payload
    except jwt.ExpiredSignatureError:
        ## Token has expired
        return None
    except jwt.InvalidTokenError:
        ## Invalid token
        return None


def token_required(view):
    @functools.wraps(view)
    def wrapped_view(**kwargs):
        auth_header = request.headers.get('Authorization')
        if not auth_header or not auth_header.startswith('Bearer '):
            return jsonify({"status": "error", "message": "Missing or invalid token"}), 401

        token = auth_header.split(' ')[1]
        payload = verify_token(token)

        if payload:
            logger.info("token accepted for %s", payload["username"])
            return view(**kwargs)
        else:
            return jsonify({"status": "error", "message": "Invalid or expired token"}), 401

    return wrapped_view



import json
import logging

from flask import Blueprint, jsonify, request

from .auth import token_required
from ..factory import get_provider

bp = Blueprint('wsgi', __name__)
logger = logging.getLogger(__name__)


@bp.route('/home')
def page_home():
    return "Base wsgi"


@bp.route('/api', methods=["GET", 'POST'])
@token_required
def page_api():
    print("dbg", dir(request))
    print("dbg args", request.args)
    print("dbg form", request.form)
    print("dbg json", request.json)
    data = "a" * 100  # open("rac5_hourly.csv").read()
    return jsonify({"status": "success", "message": "valid token", "data": data})


@bp.route('/api/<provider>', methods=['POST'])
@token_required
def page_api_provider(provider):
    logger.info("api provider %s with params %s", provider, json.dumps(request.json))
    try:
        prov_mod = get_provider(provider)
        prov = prov_mod.Provider()
    except Exception as err:
        logger.info("error provider: %s", err)
        raise err
    try:
        df, header = prov.get(**request.json)
    except Exception as err:
        logger.info("error fetch: %s", err)
        raise err
    return jsonify({"status": "success",
                    "message": "valid token",
                    "data": df.to_csv(None, sep=";"),
                    "header": header,
                    })



from datetime import UTC, datetime, timedelta

import pandas as pd
from easyprov import csv_dump, csv_header

from .fetch import fetch
from ..era5 import closest_ref
from ..provider_base import ProviderBase


class Provider(ProviderBase):

    def __init__(self):
        super().__init__()
        self.storage = self.storage_root() / "weather_archive_daily"

    def header(self):
        """Header of data stored

        Returns:
            (dict): [unit] description per columns
        """
        pth_dir = list(self.storage.glob("*/"))[0]
        pth = list(pth_dir.glob("*.csv"))[0]
        return csv_header(pth)

    def pth_storage(self, latitude, longitude, year):
        """Normalized name associated to given coordinates

        Args:
            latitude (float): [deg] latitude of place
            longitude (float): [deg] longitude of place
            year (int): year of time series

        Returns:
            (str)
        """
        return self.storage / f"{year:d}/arch_{int((latitude + 90) * 100):05d}_{int((longitude + 180) * 100):05d}.csv"

    def _ensure(self, missing, ref_lat, ref_long, date_beg, date_end, pth, df_prev=None):
        if missing == "fetch":
            df_ext, header = fetch(ref_lat, ref_long, date_beg, date_end)
            if df_prev is None:
                df = df_ext
            else:
                df = pd.concat([df_prev, df_ext])

            # write
            pth.parent.mkdir(exist_ok=True, parents=True)
            csv_dump(df, header, pth, __file__, float_format="%.2f")

            return df, header
        else:
            raise KeyError("File not available in local storage, need to fetch")

    def get(self, latitude, longitude, date_beg, date_end, closest=True, missing="raise"):
        """Get time series for given location from local storage.

        Raises: if missing="raise", KeyError if either pos or some dates have not been fetch
                previously on server

        Args:
            latitude (float): [deg] latitude of place
            longitude (float): [deg] longitude of place
            date_beg (str|datetime): First date, included
            date_end (str|datetime): End date, included
            closest (bool): Whether to interpolate from nearby locations or fetch
                            only nearest point in model
            missing (str): whether to fetch missing entry from remote server or raise error

        Returns:
            (pd.DataFrame): name of columns varies
        """
        if isinstance(date_beg, str):
            date_beg = datetime.fromisoformat(date_beg)

        if isinstance(date_end, str):
            date_end = datetime.fromisoformat(date_end)

        now = datetime.now(UTC)
        today = datetime(now.year, now.month, now.day) - timedelta(days=1)
        assert date_beg <= today
        assert date_end <= today

        assert closest

        ref_lat, ref_long = closest_ref(latitude, longitude)
        dfs = []
        header = None
        for year in range(date_beg.year, date_end.year + 1):
            pth = self.pth_storage(ref_lat, ref_long, year)
            year_end = datetime(year, 12, 31)
            if pth.exists():
                df = pd.read_csv(pth, sep=";", comment="#", parse_dates=["date"], index_col=["date"])
                if ((year < date_end.year and df.index[-1] < year_end)
                        or (year == date_end.year and df.index[-1] < date_end)):
                    # some dates were not fetched previously
                    df, header = self._ensure(missing,
                                              ref_lat,
                                              ref_long,
                                              df.index[-1] + timedelta(days=1),
                                              min(today, year_end),
                                              pth,
                                              df)
                else:
                    header = csv_header(pth)
            else:
                df, header = self._ensure(missing,
                                          ref_lat,
                                          ref_long,
                                          datetime(year, 1, 1),
                                          min(today, year_end),
                                          pth,
                                          None)

            dfs.append(df)

        df = pd.concat(dfs)

        return df.loc[date_beg: date_end], header



"""
Description of grid used by most weather models

https://confluence.ecmwf.int/display/CKB/ERA5%3A+What+is+the+spatial+reference
https://cds.climate.copernicus.eu/datasets/reanalysis-era5-single-levels?tab=overview

Grid of 0.25°x0.25° in latitude x longitude
"""

res_lat = 0.25  # [°] see header
res_long = 0.25  # [°] see header


def closest_ref(latitude, longitude):
    """Closest point on era5 grid

    Args:
        latitude (float): [deg] latitude
        longitude (float): [deg] longitude

    Returns:
        (float, float): [deg]
    """
    ref_lat = int((90 + latitude) / res_lat + 0.5) * res_lat - 90
    ref_long = int((180 + longitude) / res_long + 0.5) * res_long - 180

    return ref_lat, ref_long


def latitude_ref(latitude):
    """Latitude of reference points around latitude

    Args:
        latitude (float): [deg] latitude

    Returns:
        (float, float): [deg]
    """
    lat_min = int((90 + latitude) / res_lat) * res_lat - 90
    lat_max = int(latitude / res_lat + 1) * res_lat
    return lat_min, lat_max


def longitude_ref(longitude):
    """Longitude of reference points around longitude

    Args:
        longitude (float): [deg] longitude

    Returns:
        (float, float): [deg]
    """
    long_min = int((180 + longitude) / res_long) * res_long - 180
    long_max = int(longitude / res_long + 1) * res_long
    return long_min, long_max
